
#include "Calc.h"
#include <math.h>
#include <float.h>
#include <stdio.h>
#include <string.h>

#ifdef _WIN32
	#define isnan	_isnan
#endif //_WIN32


// Constructor
//
CCalcDouble::CCalcDouble(double init_val, CCalcObject *parent) : CCalcObject(parent)
{
	type = DOUBLE;
	value = init_val;
}


// Destructor
//
CCalcDouble::~CCalcDouble()
{

}


// Execute an instruction
//
CCalcObject *CCalcDouble::exec(CALC_OP opcode)
{
	CCalcObject *new_top = NULL;

	switch (opcode)
	{
		case DUP:
			new_top = op_dup();
			break;

		case POP:
			new_top = next;
			delete this;
			break;

		case CLR:
			if (next != NULL)
				new_top = next->exec(CLR);
			else
				new_top = NULL;
			delete this;
			break;

		case SWAP:
			if (next == NULL)
			{
				new_top = new CCalcError(getErrorMsg(1), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
			}
			else
			{
				new_top = next;
				next = next->next;
				new_top->next = this;
			}
			break;

		case CHS:
			value = -value;
			new_top = this;
			break;

		case MSTD:
			mode.format = STD;
			new_top = this;
			break;

		case MFIX:
			if ((unsigned long)(value) != value)
			{
				new_top = new CCalcError(getErrorMsg(2), this);
			}
			else
			{
				mode.decimals = (unsigned long)value;
				mode.format = FIX;
				new_top = next;
				delete this;
			}
			break;

		case MSCI:
			if ((unsigned long)value != value)
			{
				new_top = new CCalcError(getErrorMsg(2), this);
			}
			else
			{
				mode.decimals = (unsigned long)value;
				mode.format = SCI;
				new_top = next;
				delete this;
			}
			break;

		case MENG:
			if ((unsigned long)value != value)
			{
				new_top = new CCalcError(getErrorMsg(2), this);
			}
			else
			{
				mode.decimals = (unsigned long)value;
				mode.format = ENG;
				new_top = next;
				delete this;
			}
			break;

		case ADD:
			new_top = op_add();
			break;

		case SUB:
			new_top = op_sub();
			break;

		case MUL:
			new_top = op_mul();
			break;

		case DIV:
			new_top = op_div();
			break;

		case PWR:
			new_top = op_pwr();
			break;

		case ROOT:
			new_top = op_root();
			break;

		case SQRT:
			value = sqrt(value);
			new_top = this;
			break;

		case SQR:
			value *= value;
			new_top = this;
			break;

		case LOG:
			value = log10(value);
			new_top = this;
			break;

		case ALOG:
			value = pow(10, value);
			new_top = this;
			break;

		case LN:
			value = log(value);
			new_top = this;
			break;

		case EXP:
			value = exp(value);
			new_top = this;
			break;

		case SIN:
			value = sin(angle(value));
			new_top = this;
			break;

		case COS:
			value = cos(angle(value));
			new_top = this;
			break;

		case TAN:
			value = tan(angle(value));
			new_top = this;
			break;

		case ASIN:
			value = antiAngle(asin(value));
			new_top = this;
			break;

		case ACOS:
			value = antiAngle(acos(value));
			new_top = this;
			break;

		case ATAN:
			value = antiAngle(atan(value));
			new_top = this;
			break;

		case INT:
			if (value >= 0.0)
				new_top = new CCalcInteger((long)(value +0.5), next);
			else
				new_top = new CCalcInteger((long)(value -0.5), next);
			if (new_top == NULL)
			{
					error_callback("Memory allocation failure.");
					new_top = this;
			}
			else
				delete this;
			break;

		case IDV:

		case MOD:

		default:
			new_top = new CCalcError(getErrorMsg(2), this);
			break;

	}

	return new_top;
}


// Returns a string representation of the object's data value
//
const char *CCalcDouble::getString()
{
	static char repr[STRING_LEN +1];
	char format[16];
	unsigned long i;
	double mantissa;
	int exponent;

	if (isnan(value))
	{
		sprintf(repr, "");
		for (i=0; i<STRING_LEN; i++)
			strcat(repr, "-");
	}
	else
	{
		switch (mode.format)
		{
			case (STD):
				sprintf(format, "%%%dG", STRING_LEN);
				sprintf(repr, format, value);
				break;

			case (FIX):
				sprintf(format, "%%%d.%ldf", STRING_LEN, mode.decimals);
				sprintf(repr, format, value);
				break;

			case (SCI):
				sprintf(format, "%%%d.%ldE", STRING_LEN, mode.decimals);
				sprintf(repr, format, value);
				break;

			case (ENG):
				sprintf(format, "%%%d.%ldE", STRING_LEN, mode.decimals);
				mantissa = frexp(value, &exponent);
				switch (exponent % 3)
				{
					case 0:
					case 1:
					case 2:
						break;
				}
				sprintf(repr, format, value);
				break;

			default:
				error_callback("Invalid format (!AFSE).");
				sprintf(format, "%%%dG", STRING_LEN);
				sprintf(repr, format, value);
				break;
		}
	}

	return repr;
}


// Returns a single character signifying the object's type
// 
const char CCalcDouble::getType()
{
	return 'F';
}



// -----------------------------------------------------------
// OPCODE EXECUTION IMPLEMENTATION
// -----------------------------------------------------------

// Duplicates this object and places it into the stack.
//
CCalcObject *CCalcDouble::op_dup()
{
	CCalcObject *prev = new CCalcDouble(value, this);
	if (prev == NULL)
	{
		error_callback("Memory allocation failure.");
		prev = this;
	}

	return prev;
}

// Adds this object to the previous one
//
CCalcObject *CCalcDouble::op_add()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value += value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value += ((CCalcInteger *)next)->value;
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Subtracts this object from the previous one
//
CCalcObject *CCalcDouble::op_sub()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value -= value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value = ((double)((CCalcInteger *)next)->value) -value;
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Multiplies this object by the previous one
//
CCalcObject *CCalcDouble::op_mul()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value *= value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value *= ((CCalcInteger *)next)->value;
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Divides this object into the previous one
//
CCalcObject *CCalcDouble::op_div()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value /= value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value = ((double)((CCalcInteger *)next)->value)/value;
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Finds the nth power of a number
//
CCalcObject *CCalcDouble::op_pwr()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value = pow(((CCalcDouble *)next)->value, value);
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value = pow((double)((CCalcInteger *)next)->value, value);
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}

		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Finds the nth root of a number
//
CCalcObject *CCalcDouble::op_root()
{
	CCalcObject *new_top = NULL;
	CCalcObject *old_next = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value = pow(((CCalcDouble *)next)->value, 1.0/value);
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				value = pow((double)((CCalcInteger *)next)->value, 1.0/value);
				old_next = next;
				next = next->next;
				delete old_next;
				new_top = this;
				break;
			}

		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Returns the angle in radians, converting based on the value of the
// current calculator mode.
//
double CCalcDouble::angle(double in_val)
{
	switch (mode.drg)
	{
		case DEG:
			return in_val*3.1415926535/180.0;

		case RAD:
			return in_val;

		case GRAD:
			return in_val*3.1415926535/200.0;

		default:
			error_callback("Invalid angle mode (!DRG).");
			return 0.0;
	}
}


// Returns the angle in the current unit, converting based on the value of the
// current calculator mode.
//
double CCalcDouble::antiAngle(double in_val)
{
	switch (mode.drg)
	{
		case DEG:
			return in_val*180.0/3.1415926535;

		case RAD:
			return in_val;

		case GRAD:
			return in_val*200.0/3.1415926535;

		default:
			error_callback("Invalid angle mode (!DRG).");
			return 0.0;
	}
}
