
#include "Calc.h"
#include <stdio.h>
#include <string.h>


// Constructor
//
CCalcInteger::CCalcInteger(long init_val, CCalcObject *parent) : CCalcObject(parent)
{
	type = INTEGER;
	value = init_val;
}


// Destructor
//
CCalcInteger::~CCalcInteger()
{

}


// Execute an instruction
//
CCalcObject *CCalcInteger::exec(CALC_OP opcode)
{
	CCalcObject *new_top = NULL;

	switch (opcode)
	{
		case DUP:
			new_top = op_dup();
			break;

		case POP:
			new_top = next;
			delete this;
			break;

		case CLR:
			if (next != NULL)
				new_top = next->exec(CLR);
			else
				new_top = NULL;
			delete this;
			break;

		case SWAP:
			if (next == NULL)
			{
				new_top = new CCalcError(getErrorMsg(1), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
			}
			else
			{
				new_top = next;
				next = next->next;
				new_top->next = this;
			}
			break;

		case CHS:
			value = -value;
			new_top = this;
			break;

		case MSTD:
			mode.format = STD;
			new_top = this;
			break;

		case MFIX:
			mode.decimals = value;
			mode.format = FIX;
			new_top = next;
			delete this;
			break;

		case MSCI:
			mode.decimals = value;
			mode.format = SCI;
			new_top = next;
			delete this;
			break;

		case MENG:
			mode.decimals = value;
			mode.format = ENG;
			new_top = next;
			delete this;
			break;

		case ADD:
			new_top = op_add();
			break;

		case SUB:
			new_top = op_sub();
			break;

		case MUL:
			new_top = op_mul();
			break;

		case DIV:
			new_top = op_div();
			break;

		case PWR:
		case ROOT:
		case SQRT:
		case SQR:
		case LOG:
		case ALOG:
		case LN:
		case EXP:
		case SIN:
		case COS:
		case TAN:
		case ASIN:
		case ACOS:
		case ATAN:
			new_top = doubleFunc(opcode);
			break;

		case INT:
			new_top = this;
			break;

		case BSL:
		case BSR:
		case ASL:
		case ASR:
			new_top = intFunc(opcode);
			break;

		case CMP:
			value = -value;
			new_top = this;
			break;

		case NOT:
			value = ~value;
			new_top = this;
			break;

		case AND:
		case OR:
		case XOR:
		case EQU:
			new_top = intFunc(opcode);
			break;

		case IDV:

		case MOD:

		default:
			new_top = new CCalcError(getErrorMsg(2), this);
			if (new_top == NULL)
			{
				error_callback("Memory allocation failure.");
				new_top = this;
			}
			break;

	}

	return new_top;
}


// Returns a string representation of the object's data value
//
const char *CCalcInteger::getString()
{
	static char repr[STRING_LEN +1];
	char format[16];
	unsigned long i;
	unsigned long stopn;
	unsigned long temp;

	switch (mode.base)
	{
		case 2:
			sprintf(repr, "");
			temp = (unsigned long)value;
			stopn = STRING_LEN < mode.bindigits ? STRING_LEN : mode.bindigits;
			if (stopn < mode.bindigits)
				for (i=0; i<STRING_LEN; i++)
					strcat(repr, "*");
			else
			{
				temp <<= (32 -mode.bindigits);
				for (i=0; i<mode.bindigits; i++)
				{
					if (temp & 0x80000000)
						strcat(repr, "1");
					else
						strcat(repr, "0");
					temp <<= 1;
				}
			}
			break;

		case 8:
			sprintf(format, "%%%dlo", STRING_LEN);
			sprintf(repr, format, (unsigned long) value);
			break;

		case 10:
			if (mode.sign10)
			{
				sprintf(format, "%%%dld", STRING_LEN);
				sprintf(repr, format, value);
			}
			else
			{
				sprintf(format, "%%%dlu", STRING_LEN);
				sprintf(repr, format, (unsigned long) value);
			}
			break;

		case 16:
			sprintf(format, "%%%dlX", STRING_LEN);
			sprintf(repr, format, value);
			break;

		default:
			error_callback("Invalid base (!BODH).");
			sprintf(format, "%%%dld", STRING_LEN);
			sprintf(repr, format, value);
	}

	return repr;
}


// Returns a single character signifying the object's type
//
const char CCalcInteger::getType()
{
	return 'I';
}


// -----------------------------------------------------------
// OPCODE EXECUTION IMPLEMENTATION
// -----------------------------------------------------------

// Duplicates this object and places it into the stack.
//
CCalcObject *CCalcInteger::op_dup()
{
	CCalcObject *prev = new CCalcInteger(value, this);
	if (prev == NULL)
	{
		error_callback("Memory allocation failure.");
		prev = this;
	}

	return prev;
}

// Adds this object to the previous one
//
CCalcObject *CCalcInteger::op_add()
{
	CCalcObject *new_top = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value += value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				((CCalcInteger *)next)->value += value;
				new_top = next;
				delete this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Subtracts this object from the previous one
//
CCalcObject *CCalcInteger::op_sub()
{
	CCalcObject *new_top = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value -= value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				((CCalcInteger *)next)->value -= value;
				new_top = next;
				delete this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Multiplies this object by the previous one
//
CCalcObject *CCalcInteger::op_mul()
{
	CCalcObject *new_top = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value *= value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				((CCalcInteger *)next)->value *= value;
				new_top = next;
				delete this;
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// Divides this object into the previous one
//
CCalcObject *CCalcInteger::op_div()
{
	CCalcObject *new_top = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (next->type)
	{
		case (DOUBLE):
			{
				((CCalcDouble *)next)->value /= (double)value;
				new_top = next;
				delete this;
				break;
			}

		case (INTEGER):
			{
				new_top = new CCalcDouble(((double)((CCalcInteger *)next)->value)/((double)value), next->next);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
					
				}
				else
				{
					delete next;
					delete this;
				}
				break;
			}
		default:
			{
				new_top = new CCalcError(getErrorMsg(0), this);
				if (new_top == NULL)
				{
					error_callback("Memory allocation failure.");
					new_top = this;
				}
				break;
			}
	}

	return new_top;
}


// This function handles a variety of opcodes.  It first converts
// the current INTEGER data type to a DOUBLE, then executes the
// opcode on the DOUBLE.  Used to execute floating-point calculations
// on INTEGERs.
//
CCalcObject *CCalcInteger::doubleFunc(CALC_OP opcode)
{
	CCalcObject *new_top = NULL;

	// Allocate a new DOUBLE
	//
	new_top = new CCalcDouble((double)value, next);
	if (new_top == NULL)
	{
		error_callback("Memory allocation failure.");
		return this;
	}

	// Pass on the execution to the new object
	//
	new_top = new_top->exec(opcode);
	delete this;
	return new_top;
}


// This function handles integer function opcodes that require two
// parameters from the stack.
//
CCalcObject *CCalcInteger::intFunc(CALC_OP opcode)
{
	CCalcObject *new_top = NULL;

	if (next == NULL)
	{
		new_top = new CCalcError(getErrorMsg(1), this);
		if (new_top == NULL)
		{
			error_callback("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	if (next->type != INTEGER)
	{
		new_top = new CCalcError(getErrorMsg(0), this);
		if (new_top == NULL)
		{
			(*error_callback)("Memory allocation failure.");
			new_top = this;
		}
		return new_top;
	}

	switch (opcode)
	{
		case BSL:
			((CCalcInteger *)next)->value = ((CCalcInteger *)next)->value << value;
			new_top = next;
			delete this;
			break;

		case BSR:
			((CCalcInteger *)next)->value = ((CCalcInteger *)next)->value >> value;
			new_top = next;
			delete this;
			break;

		case ASL:
			((CCalcInteger *)next)->value = ((CCalcInteger *)next)->value << value;
			new_top = next;
			delete this;
			break;

		case ASR:
			((CCalcInteger *)next)->value = ((CCalcInteger *)next)->value << value;
			new_top = next;
			delete this;
			break;

		case AND:
			((CCalcInteger *)next)->value &= value;
			new_top = next;
			delete this;
			break;

		case OR:
			((CCalcInteger *)next)->value |= value;
			new_top = next;
			delete this;
			break;

		case XOR:
			((CCalcInteger *)next)->value ^= value;
			new_top = next;
			delete this;
			break;

		case EQU:
			((CCalcInteger *)next)->value = ~(((CCalcInteger *)next)->value ^ value);
			new_top = next;
			delete this;
			break;

		default:
			error_callback("Invalid opcode passed to INTEGER.");
			delete new_top;
			new_top = this;
			break;
	}

	return new_top;
}
