#ifndef __CALC_OBJECT_H
#define __CALC_OBJECT_H


// These define the various object types
//
enum CALC_OBJECT_DESCR
{
	UNKNOWN,
	DOUBLE,
	INTEGER,
	STRING,
	ERROR
};


// These define all the opcodes.  "2", "1", etc. in the comments indicate
// the object at that stack level.
//
typedef enum CALC_OP
{
	// stack operations
	//
	DUP,		// duplicate object
	POP,		// pop from stack
	CLR,		// clear the stack
	SWAP,		// swap the top two objects
	SWPN,		// swap the top with the nth item
	CHS,		// change sign of top item

	// modes
	//
	MSTD,
	MFIX,
	MSCI,
	MENG,

	// normal math
	//
	ADD,		// 2 + 1
	SUB,		// 2 - 1
	MUL,		// 2 * 1
	DIV,		// 2 / 1
	PWR,		// 2 ^ 1
	ROOT,		// 1st root of 2
	SQRT,		// sqrt(1)
	SQR,		// 1*1
	LOG,		// log(10) of 1
	ALOG,		// 10^1
	LN,			// log(e) of 1
	EXP,		// e^1
	SIN,		// sin(1)
	COS,		// cos(1)
	TAN,		// tan(1)
	ASIN,		// asin(1)
	ACOS,		// acos(1)
	ATAN,		// atan(1)

	// binary arithmetic
	//
	INT,		// convert 1 to INTEGER
	BSL,		// bit shift left (fill with zeros)
	BSR,		// bit shift right (fill with zeros)
	ASL,		// arithmetic shift left (fill with zeros)
	ASR,		// arithmetic shift right (fill with sign)
	CMP,		// 2's complement of 1
	NOT,		// binary inverse of 1 (also 1's complement of 1)
	AND,		// 1 AND 2
	OR,			// 1 OR 2
	XOR,		// 1 XOR 2
	EQU,		// 1 EQU 2
	IDV,		// integer part of (2 / 1)
	MOD,		// remainder part of (2 / 1)

	// memory operations
	//
	STO,		// store 2 into variable named 1
	RCL,		// recall variable named 1
	DIR,		// list all variables in memory

	// end of table marker
	//
	NOP
};


// This defines a lookup table to get string names for each opcode
//
typedef struct
{
	CALC_OP		opcode;
	char		name[5];
} OPCODE_TAB;


extern OPCODE_TAB opcode_table[];
extern const char *getName(CALC_OP opcode);


// These define how long string representation of objects are
//
#define STRING_LEN			32


// This defines an error string table
//
#define ERROR_MSG_LEN		32

typedef struct
{
	char	msg[ERROR_MSG_LEN +1];
} ERROR_STR;

extern ERROR_STR error_string[];

extern const char *getErrorMsg(long index);


// Angle mode
//
typedef enum DRG_MODE
{
	DEG,
	RAD,
	GRAD
};

// Floating point display
//
typedef enum FLOAT_MODE
{
	STD,
	FIX,
	SCI,
	ENG
};


// This defines the calculator modes
//
typedef struct
{
	DRG_MODE		drg;
	FLOAT_MODE		format;
	unsigned long	decimals;
	unsigned long	base;
	unsigned long	bindigits;
	bool			sign10;
} CALC_MODES;


extern CALC_MODES mode;
extern void (*error_callback)(const char *msg);


// class CCalcObject
//
// This is the base class for all types of objects that can be placed on the
// calculator stack.  This is a virtual class, and cannot be instantiated.
//
// Matthew J. Gelhaus
// March 12, 2004
//
class CCalcObject  
{
	public:
		CCalcObject(CCalcObject *parent=0);
		virtual ~CCalcObject();

		virtual CCalcObject *exec(CALC_OP opcode) = 0;
		virtual const char *getString() = 0;
		virtual const char getType() = 0;

		CCalcObject *next;
		CALC_OBJECT_DESCR type;
};


#endif //__CALC_OBJECT_H

